REQUIRE avlTree ~nn/lib/utils/avltree.f
REQUIRE ICOMPARE ~nn/lib/wcmatch.f
REQUIRE ?FREE ~nn/lib/free.f
REQUIRE FF-ATTRIB ~nn/lib/find.f
REQUIRE MD5File ~nn/lib/security/md5dll.f
REQUIRE <TIB ~nn/lib/tib.f
REQUIRE get-string ~nn/lib/getstr.f
REQUIRE S>NUM ~nn/lib/s2num.f
REQUIRE [NONAME ~nn/lib/noname.f
REQUIRE INCLUDE-WITH ~nn/lib/includewith.f
REQUIRE MAKE-FILE ~nn/lib/file.f
REQUIRE N>S ~nn/lib/num2s.f

USER lastFileNode   \ for filling fields

CLASS: fileObject
    var  vFileAttributes
    dvar vCreationTime
    dvar vLastAccessTime
    dvar vLastWriteTime 
    dvar vFileSize
    var vMD5    \ pointer to ascii string of MD5
    var vFullPath
    var vLen
    var vFlags

    M: setName ( a u -- ) DUP vLen ! S>ZALLOC vFullPath ! ;

    CONSTR: init ( a u -- ) setName ;

    DESTR: free vFullPath @ ?FREE ;

    M: setAttr vFileAttributes ! ;
    M: setCTime ( ft --) vCreationTime 2! ;
    M: setWTime vLastWriteTime 2! ;
    M: setATime vLastAccessTime 2! ;
    M: setSize  vFileSize 2! ;
    M: setMD5 ?DUP IF S>ZALLOC vMD5 !  ELSE DROP THEN ;
    M: setFlags vFlags ! ;

    M: getName  vFullPath @ vLen @ ;    
    M: getMD5 vMD5 @ ?DUP IF ASCIIZ> ELSE S" 0" THEN ;
    M: getAttr vFileAttributes @ ;
    M: getCTime ( ft --) vCreationTime 2@ ;
    M: getWTime vLastWriteTime 2@ ;
    M: getATime vLastAccessTime 2@ ;
    M: getSize vFileSize 2@ ;
;CLASS

WITH fileObject
CLASS: fileNode <SUPER avlNode
    var vFileObj
    CONSTR: init init vFileObj ! ;
    DESTR: free vFileObj @ DELETE ;

    VM: toString vFileObj @ \ ." toString=" DUP . CR
        => getName ;
    VM: Compare ( obj -- n ) \ ." Compare=" DUP . CR
        => getName toString ICOMPARE ;
;CLASS

CLASS: fileList <SUPER avlTree
    var vLastFile
    var vFH
    var vMD5flag
    var vBuf
    var vBufSize
    16 chars vMD5buf
    
    M: lastFile vLastFile @ ;
    CONSTR: init fileNode init 1024 vBufSize ! 
          vBufSize @ ALLOCATE THROW vBuf !
    ;
    DESTR: free free vBuf @ ?FREE ;

    M: Insert ( a u -- ) 
        fileObject NEW vLastFile !
        \ ." lastFile=" lastFile . CR
        lastFile Insert ;

    M: setInfo ( flags attr dCtime dAtime dWtime dsize amd5 umd5 -- )
        lastFile => setMD5
        lastFile => setSize     lastFile => setWTime   
        lastFile => setATime  lastFile => setCTime
        lastFile => setAttr     lastFile => setFlags 
        ;
    M: calcMD5 ( -- a u)
        vMD5buf lastFile => getName MD5File THROW
        vMD5buf MD5String 
    ;

    M: InsertFF ( a u -- )
        Insert
        0 FF-ATTRIB FF-CREATION-TIME  FF-WRITE-TIME FF-ACCESS-TIME FF-SIZE
        vMD5flag @ IF calcMD5 ELSE S" " THEN
        setInfo
    ;

\ flags attribs ctime wtime size md5 path
    M: InsertLine ( a u -- )
        <TIB
            get-string S>NUM       \ flags
            get-string S>NUM       \ file attributes
            get-string SDD.MM.YYYY/hh:mm:ss YMDHMS>FT   \ creation time
            get-string SDD.MM.YYYY/hh:mm:ss YMDHMS>FT   \ write time
            0.                                          \ access time
            get-string S>DOUBLE     \ size
            get-string              \ md5sum
            get-string              \ filename
        TIB>
        Insert
        setInfo
    ;
    
    VM: write-line ( a u -- ) vFH @ WRITE-LINE THROW ;
    VM: wr ( a u -- ) vFH @ WRITE-FILE THROW ;
    VM: openw ( a u -- ) R/W MAKE-FILE THROW vFH ! ;
    VM: openr ( a u -- ) R/O OPEN-FILE THROW vFH ! ;
    VM: close vFH @ ?DUP IF CLOSE-FILE THROW THEN ;
    VM: read-line ( a u -- u1 ?) vFH @ READ-LINE THROW ;
    : rl vBuf @ vBufSize @ read-line ;
    : ch-wr SP@ 1 wr DROP ;
    : bl-wr BL ch-wr ;
    : wr+bl wr bl-wr ;
    
    M: readFromFile ( a u -- )
        openr
        BEGIN  vBuf @ vBufSize @ read-line WHILE
           vBuf @ SWAP \ 2DUP TYPE CR
           InsertLine
        REPEAT
        DROP
        close
    ;
     
\ flags attribs ctime wtime size md5 path
    
    M: writeToFile ( a u -- )
        openw
        [NONAME ( node )
            ->CLASS fileNode vFileObj @ vLastFile !
            S" 0" wr+bl
            lastFile => getAttr N>S wr+bl
            lastFile => getCTime FT>DD.MM.YYYY/hh:mm:ss wr+bl
            lastFile => getWTime FT>DD.MM.YYYY/hh:mm:ss wr+bl
            lastFile => getSize DOUBLE>S wr+bl
            lastFile => getMD5 wr+bl
            [CHAR] " ch-wr
            lastFile => getName wr
            [CHAR] " ch-wr
            13 ch-wr 10 ch-wr
        NONAME] endOrder
        close
    ;
;CLASS
ENDWITH

REQUIRE gzopen ~nn/lib/zlib.f

CLASS: gzfileList <SUPER fileList
    CONSTR: init init ;
    DESTR: free free ;
    VM: wr ( a u -- ) SWAP vFH @ gzwrite vFH @ ?GZERR ;
    VM: write-line ( a u -- ) wr LT LTL @ wr ;
    : gzopen1 ( a u a-mode u-mode -- h) 
        DROP NIP SWAP gzopen 
        ?DUP 0= IF GetLastError THROW 0 THEN  vFH ! ;
    VM: openw ( a u -- ) S" wb5" gzopen1 ;
    VM: openr ( a u -- ) S" rb" gzopen1 ;
    VM: close vFH @ ?DUP IF gzclose DROP THEN ;
    VM: read-line ( a u -- u1 ?) SWAP vFH @ gzgets ?DUP vFH @ ?GZERR 
        ASCIIZ> NIP TRUE
    ;
;CLASS


REQUIRE \EOF ~nn/lib/eof.f
\ EOF
~nn/lib/utils/avlprint.f
gzfileList POINTER fl
: test1
  gzfileList NEW TO fl
  fl vMD5flag ON
  RECURSIVE
  S" *.*" FOR-FILES
    IS-DIR? 0=
    IF
     FOUND-RELPATH TYPE CR
     FOUND-RELPATH fl InsertFF 
    THEN
  ;FOR-FILES
  S" testfl.gz" fl writeToFile
;
: test2
  gzfileList NEW TO fl
  fl vMD5flag ON
  S" testfl.gz" fl readFromFile
  fl print1
;
\ test1
test2
