REQUIRE CLASS: ~nn/class/class.f
REQUIRE PARSE-MASKS ~nn/lib/masks.f
\ REQUIRE AddNodeSorted ~nn/lib/zlist.f
REQUIRE FOR-FILES ~nn/lib/for-files.f
\ REQUIRE FIND-FIRST-FILE ~nn/lib/find.f
REQUIRE S+ ~nn/lib/az.f
REQUIRE ICOMPARE ~nn/lib/wcmatch.f
REQUIRE GetFileSA ~nn/lib/security/secinfo.f
REQUIRE MAKE-DIR ~nn/lib/makedirs.f
REQUIRE COPY-FILE ~nn/lib/copyfile.f

CLASS: FileNode
   RECORD: FIND_DATA
    var  dwFileAttributes; 
    dvar ftCreationTime; 
    dvar ftLastAccessTime; 
    dvar ftLastWriteTime; 
    var  nFileSizeHigh; 
    var  nFileSizeLow; 
    var  dwReserved0; 
    var  dwReserved1; 
    260 FIELD cFileName
    14  FIELD cAlternateFileName
   ;RECORD /FIND_DATA
    var vFullPath

    CONSTR: init ( a u -- ) S>ZALLOC vFullPath ! ;
    DESTR: free vFullPath @ ?DUP IF FREE DROP THEN ;
    M: FIND_DATA! FIND_DATA /FIND_DATA CMOVE ;
;CLASS

CLASS: FileList
    var vIncMasks     \ list of included masks
    var vExcMasks     \ list of excluded masks
    var vList
    var vLastNode
    var vCnt

    CONSTR: init ;
    : FreeFileNode NodeValue DELETE ;
    DESTR: free
        ['] FreeFileNode vList DoList
        vList FreeList ;

    M: AddMasks   ( a u --)  vIncMasks PARSE-MASKS  ;
    M: AddMask    ( a u --)  vIncMasks ADD-MASK ;
    M: AddXMasks  ( a u --)  vExcMasks PARSE-MASKS  ;
    M: AddXMask   ( a u --)  vExcMasks ADD-MASK ;
    M: CheckPath ( a u -- ?)
        vIncMasks @ IF 2DUP vIncMasks FIND-MASK ELSE TRUE THEN >R
        vExcMasks @ IF vExcMasks FIND-MASK ELSE 2DROP FALSE THEN 0=
        R> AND ;
    VM: Convert ( a u -- a u)  ;
    M: FindGreat ( -- list)
        vList
        BEGIN DUP @ ?DUP WHILE
            SWAP >R 
            NodeValue ->CLASS FileNode vFullPath @ ASCIIZ> 2OVER ICOMPARE 0 >
            IF R> EXIT THEN
            R> @
        REPEAT ;

    M: AddNodeSorted ( file-node -- ) DUP vLastNode ! FindGreat AddNode ;

    M: AddPath ( a u -- ?)
        Convert
        2DUP CheckPath 
        IF  FileNode NEW AddNodeSorted
            TRUE
            vCnt 1+!
        ELSE
            2DROP FALSE
        THEN
    ;
    VM: MakeList ( a u -- )  ;
    : .FileNode NodeValue ->CLASS FileNode vFullPath @ ASCIIZ> TYPE CR ;
    M: .List  ['] .FileNode vList DoList ;
;CLASS

CLASS: WinFileList <SUPER FileList
    var vSrcPath
    var vDstPath
    var vCopyAccessRights
    var vSrcBuf
    var vDstBuf
    var vSubtree
    var vBaseLen

    M: Src vSrcPath @ ASCIIZ> ;
    M: Dst vDstPath @ ASCIIZ> ;
    : DIR+ >R R@ ZPLACE S" \" R@ +ZPLACE R@ +ZPLACE  R> ASCIIZ> ;
    M: SRC+ vSrcBuf @ DIR+ ;
    M: DST+ vDstBuf @ DIR+ ;

    CONSTR: init init 
        1024 ALLOCATE THROW vSrcBuf !
        1024 ALLOCATE THROW vDstBuf ! ;

    DESTR: free free 
        vSrcBuf @ ?FREE vDstBuf @ ?FREE
        vSrcPath @ ?FREE vDstPath @ ?FREE ;

    : ABS-PATH ( a u len -- a1 u1 )
        vBaseLen @ + vSrcBuf @ + >R
        S" \" R@ ZPLACE
        R> 1+ ZPLACE vSrcBuf @ ASCIIZ>  ;

    : REL-PATH ( a u len -- a1 u1 )
        ABS-PATH 
        >R vBaseLen @ + 1+
        R> vBaseLen @ - 1- ;

    USER-VALUE len
    USER-VALUE cnt
    : MAKE-TREE-LIST ( len \ cnt -- file-cnt )
        len >R cnt >R  TO len
        0 TO cnt
        ~RECURSIVE
        S" *.*" len ABS-PATH    \ 2DUP TYPE ."  <----" CR
        FOR-FILES
            FOUND-FILENAME len REL-PATH TYPE
            IS-DIR? 0= 0= vSubtree @ AND
            IF
\                ."  is dir" CR
                IF
                    FOUND-FILENAME NIP len + 1+
                    RECURSE
                    cnt + TO cnt
                THEN
            ELSE
\                ."  is file" CR
                cnt 1+ TO cnt
            THEN
            KEY DROP
        ;FOR-FILES
        cnt DUP . CR
        R> TO cnt R> TO len
    ;

    VM: MakeList ( a u -- )
        DUP vBaseLen ! vSrcBuf @ ZPLACE
        0 MAKE-TREE-LIST
    ;

    : SrcSA 
        vCopyAccessRights @
        IF FOUND-RELPATH SRC+ GetDirSA ELSE 0 THEN ;

    M: Copy ( a-from u1 a-to u2 -- )
        S>ZALLOC vDstPath ! S>ZALLOC vSrcPath !

        vCopyAccessRights @ IF Src GetDirSA ELSE 0 THEN MAKE-DIR-SA !
        Src MAKE-DIRS THROW
         
    ;
;CLASS

